#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Convenience function to run the app directly."""

from multiprocessing.sharedctypes import Synchronized as Value
from typing import Dict, Sequence, Tuple

from .app import App
from .knob import Knob, T


def run(tui: Dict[str, Tuple[Value, Sequence[T]]]) -> None:
    """Convenience function to run the app directly from the valmix module.

    Args:
        tui: Dictionary defining the terminal user interface.

    A TUI dictionary looks like this:

        {
            "foo": (mp.Value("f", 10.0), np.arange(0.0, 10.0, 0.5)),
            "bar": (mp.Value("i", 2), range(-7, 7)),
        }

    Each key defines a parameter whose value to tune. Each value (of the
    key-value dictionary) is a pair consisting of (1) a multiprocessing Value
    instance, holding the actual parameter to tune, and (2) the values the
    parameter may take, typically a range for integers or a NumPy linspace for
    floating-point numbers.
    """
    knobs = [Knob(key, value, values) for key, (value, values) in tui.items()]
    App(knobs).run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Adjust two numerical values from a terminal user interface.

The value "foo" is an integer in a custom range, while the value "bar" is a
floating point number living in a custom NumPy array range.
"""

import multiprocessing as mp

import numpy as np

import valmix

if __name__ == "__main__":
    valmix.run({
        "foo": (mp.Value("i", 0), range(-10, 10, 3)),
        "bar": (mp.Value("f", 1.0), np.arange(-1.0, 3.0, 0.1)),
    })



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Knob interface."""

from multiprocessing.sharedctypes import Synchronized as Value
from typing import Generic, Sequence, TypeVar

T = TypeVar("T")


class Knob(Generic[T]):
    r"""Name and list values for a target quantity.

    Attributes:
        name: Display name of the value.
        value: Actual multiprocessing synchronized shared object.
        values: Values the internal value may take.
    """

    def __init__(
        self,
        name: str,
        value: Value,
        values: Sequence[T],
    ) -> None:
        """Create a new knob.

        Args:
            name: Display name of the value.
            value: Actual multiprocessing synchronized shared object.
            values: Values the internal value may take.
        """
        nb_values = len(values)
        initial_value = value.value
        self.__index = nb_values // 2
        self.__mp_value = value
        self.__name = name
        self.__nb_values = nb_values
        self.__values = sorted(values)
        #
        self.__snap_to(initial_value)
        self.__update_mp_value()

    def __snap_to(self, snap_value: T) -> None:
        """Snap to available value closest to a given one.

        Args:
            snap_value: Quantity to get close to using values from the set.
        """
        self.__index = 0
        best_dist = abs(self.__values[0] - snap_value)
        for i in range(1, self.__nb_values):
            cur_dist = abs(self.__values[i] - snap_value)
            if cur_dist < best_dist:
                self.__index = i
                best_dist = cur_dist

    def __update_mp_value(self) -> None:
        """Update internal multiprocessing value from current index."""
        new_value = self.__values[self.__index]
        with self.__mp_value.get_lock():
            self.__mp_value.value = new_value

    def advance(self, step: int) -> None:
        """Advance to the next value listed when configuring the knob.

        Args:
            step: Number of steps to add to the list index.
        """
        self.__index += step
        if self.__index >= self.__nb_values:
            self.__index = self.__nb_values - 1
        elif self.__index < 0:
            self.__index = 0
        self.__update_mp_value()

    @property
    def current_index(self) -> int:
        """Get current index in the list of available values.

        Returns:
            Current index of the knob.
        """
        return self.__index

    @property
    def current_value(self) -> T:
        """Get current value of the knob.

        Returns:
            Current value of the knob.
        """
        return self.__mp_value.value

    @property
    def name(self) -> str:
        """Knob name to display."""
        return self.__name

    @property
    def nb_values(self) -> int:
        """Get the number of different values that the knob can take.

        Returns:
            Number of different values that the knob can take.
        """
        return self.__nb_values

    def __repr__(self) -> str:
        """Get a string representation of the knob's value.

        Returns:
            String representation of the current value of the knob.
        """
        v = self.current_value
        if isinstance(v, float):
            return f"{v:.2f}" if 1e-2 <= abs(v) < 100.0 else f"{v:.1e}"
        return str(self.current_value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Main application screen."""

from typing import List

import textual.app
from textual.app import ComposeResult
from textual.binding import Binding
from textual.widgets import Footer, Header

from .knob import Knob
from .knob_widget import KnobWidget


class App(textual.app.App):
    """Valmix application.

    Attributes:
        knobs: Knobs manipulated via the app interface.
    """

    TITLE = "Valmix"

    BINDINGS = [
        Binding(key="down,j", action="next", description="Next"),
        Binding(key="up,k", action="previous", description="Previous"),
        Binding(key="left,h", action="decrease", description="Decrease"),
        Binding(key="right,l", action="increase", description="Increase"),
        Binding(key="q", action="quit", description="Quit"),
    ]

    def __init__(self, knobs: List[Knob]) -> None:
        """Initialize application.

        Args:
            knobs: Knobs manipulated via the app interface.
        """
        self.knobs = {knob.name: knob for knob in knobs}
        super().__init__()

    def compose(self) -> ComposeResult:
        """Compose the application screen."""
        yield Header()
        for knob in self.knobs.values():
            yield KnobWidget(knob)
        yield Footer()

    def action_decrease(self) -> None:
        """Decrease currently-selected knob."""
        if self.focused is not None:
            knob = self.query_one(f"#{self.focused.id}-knob")
            if isinstance(knob, KnobWidget):  # should be the case
                knob.advance(-1)

    def action_increase(self) -> None:
        """Increase currently-selected knob."""
        if self.focused is not None:
            knob = self.query_one(f"#{self.focused.id}-knob")
            if isinstance(knob, KnobWidget):  # should be the case
                knob.advance(+1)

    def action_previous(self) -> None:
        """Focus previous knob."""
        self.screen.focus_previous()

    def action_next(self) -> None:
        """Focus next knob."""
        self.screen.focus_next()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""KnobWidget widget definition."""

from textual.app import ComposeResult
from textual.containers import Widget
from textual.widgets import Button, Label, ProgressBar

from .knob import Knob


class KnobWidget(Widget):
    """Widget combining a focus button, a progress bar and a value label.

    Attributes:
        knob: Knob associated with the widget.
    """

    DEFAULT_CSS = r"""

    KnobWidget {
        height: 3;
        layout: horizontal;
    }

    KnobWidget Button,
    KnobWidget Button:enabled, {
        border: wide $primary;
        margin: 0 2;
        width: 3;
    }

    KnobWidget Button:focus {
        background: $accent;
    }

    KnobWidget ProgressBar {
        margin: 1 3;
        width: 6fr;
    }

    KnobWidget Bar {
        width: 1fr;
    }

    KnobWidget PercentageStatus {
        display: none;
    }

    KnobWidget Label {
        margin: 1 2;
        width: 8;
    }

    """

    knob: Knob
    label: Label
    progress_bar: ProgressBar

    def __init__(self, knob: Knob):
        """Create a new widget.

        Args:
            knob: Knob associated with the widget.
        """
        self.knob = knob
        self.label = Label()  # updated by compose()
        self.progress_bar = ProgressBar()  # updated by compose()
        super().__init__(id=f"{knob.name}-knob")

    def compose(self) -> ComposeResult:
        """Render the widget."""
        yield Button(self.knob.name, id=self.knob.name)

        self.progress_bar = ProgressBar(
            total=self.knob.nb_values - 1,
            show_eta=False,
            show_percentage=True,
            id=f"{self.knob.name}-bar",
        )
        self.progress_bar.advance(self.knob.current_index)
        yield self.progress_bar

        self.label = Label(repr(self.knob), id=f"{self.knob.name}-value")
        yield self.label

    def advance(self, step: int):
        """Advance the underlying knob by a number of steps.

        Args:
            step: Number of steps to add to advance the knob by.
        """
        previous_index = self.knob.current_index
        self.knob.advance(step)
        if self.knob.current_index == previous_index:
            return
        self.progress_bar.advance(step)  # can advance beyond its total
        self.label.update(repr(self.knob))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Tune a linear function from the command line."""

import multiprocessing as mp
import time

import numpy as np

import valmix


def main(kp: mp.Value, kd: mp.Value):
    """A dummy program we want to tune.

    Args:
        kp: First tuning parameter.
        kd: Second tuning parameter.
    """
    with open("/tmp/output", "w") as output:
        for _ in range(15):
            u = np.clip(kp.value * 1.0 + kd.value * 0.1, 5.0, 20.0)
            output.write(f"{u}\n")
            output.flush()
            time.sleep(1.0)


if __name__ == "__main__":
    kp = mp.Value("f", 10.0)
    kd = mp.Value("f", 1.0)

    # Call the main function in a separate process
    main_process = mp.Process(target=main, args=(kp, kd))
    main_process.start()

    # Display the terminal user interface in this process (blocking call)
    valmix.run(
        {
            "kp": (kp, np.arange(0.0, 20.0, 0.5)),
            "kd": (kd, np.arange(0.0, 10.0, 0.5)),
        }
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Unit tests for the Knob interface."""

import multiprocessing as mp
import unittest

import numpy as np

from valmix import Knob


class TestKnob(unittest.TestCase):
    """Tests the Knob interface."""

    def test_available_initial_value(self):
        v = mp.Value("i", 12)
        knob = Knob("v", v, range(5, 15, 1))
        self.assertEqual(knob.current_value, v.value)

    def test_out_of_range_initial_value(self):
        v = mp.Value("i", 12)
        knob = Knob("v", v, range(5, 15, 3))
        self.assertEqual(knob.current_value, 11)
        self.assertEqual(v.value, 11)

    def test_advance(self):
        v = mp.Value("f", 12.1)
        knob = Knob("v", v, np.arange(5, 15, 1.0))
        self.assertAlmostEqual(knob.current_value, 12.0)
        self.assertAlmostEqual(v.value, 12.0)
        knob.advance(+1)
        self.assertAlmostEqual(knob.current_value, 13.0)
        self.assertAlmostEqual(v.value, 13.0)
        knob.advance(-1)
        knob.advance(-1)
        self.assertAlmostEqual(knob.current_value, 11.0)
        self.assertAlmostEqual(v.value, 11.0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Adjust numerical values from a terminal user interface."""

from .app import App
from .knob import Knob
from .run import run

__version__ = "0.0.1"

__all__ = [
    "App",
    "Knob",
    "run",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Unit tests for the overall app."""

import multiprocessing as mp
import unittest

import numpy as np

import valmix


class TestApp(unittest.IsolatedAsyncioTestCase):
    """Tests for valmix.App."""

    async def test_run(self):
        foo = valmix.Knob("foo", mp.Value("i"), range(-10, 10, 3))
        bar = valmix.Knob("baz", mp.Value("f"), np.arange(-1.0, 3.0, 0.1))
        app = valmix.App([foo, bar])
        async with app.run_test() as pilot:
            await pilot.click("#foo")
            self.assertEqual(app.focused.id, "foo")
            await pilot.click("#baz")
            self.assertEqual(app.focused.id, "baz")
